#include "config.h"

#include <glib.h>
#include <unistd.h>

#include "boilerplate-log.h"

static int log_verbosity;

/**
 * log_level_to_string:
 * @log_level: A #GLogLevelFlags.
 *
 * The string name for the G_LOG_LEVEL_*
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static const gchar *
log_level_to_string (GLogLevelFlags log_level)
{
    switch (((gulong)log_level & G_LOG_LEVEL_MASK))
    {
        case G_LOG_LEVEL_ERROR:    return "   ERROR";
        case G_LOG_LEVEL_CRITICAL: return "CRITICAL";
        case G_LOG_LEVEL_WARNING:  return " WARNING";
        case G_LOG_LEVEL_MESSAGE:  return " MESSAGE";
        case G_LOG_LEVEL_INFO:     return "    INFO";
        case G_LOG_LEVEL_DEBUG:    return "   DEBUG";

        default:
          return " UNKNOWN";
    }
}

/**
 * log_level_to_string_with_color:
 * @log_level: A #GLogLevelFlags.
 *
 * The string name for the G_LOG_LEVEL_* but with color
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static const gchar *
log_level_to_string_with_color (GLogLevelFlags log_level)
{
    switch (((gulong)log_level & G_LOG_LEVEL_MASK))
    {
        case G_LOG_LEVEL_ERROR:    return "   \033[1;31mERROR\033[0m";
        case G_LOG_LEVEL_CRITICAL: return "\033[1;35mCRITICAL\033[0m";
        case G_LOG_LEVEL_WARNING:  return " \033[1;33mWARNING\033[0m";
        case G_LOG_LEVEL_MESSAGE:  return " \033[1;32mMESSAGE\033[0m";
        case G_LOG_LEVEL_INFO:     return "    \033[1;32mINFO\033[0m";
        case G_LOG_LEVEL_DEBUG:    return "   \033[1;32mDEBUG\033[0m";

        default:
            return " UNKNOWN";
    }
}

/**
 * log_handler:
 * @log_domain: A string containing the log section.
 * @log_level: A #GLogLevelFlags.
 * @message: The log message.
 * @user_data: User data supplied to g_log_set_default_handler().
 *
 * Default log handler that will dispatch log messages to configured logging
 * destinations.
 *
 * The log_verbosity level must be high enough to see the following messages:
 * lvl 1 -v   G_LOG_LEVEL_MESSAGE >
 * lvl 2 -vv  G_LOG_LEVEL_INFO >
 * lvl 3 -vvv G_LOG_LEVEL_DEBUG >
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
log_handler (const gchar    *log_domain,
             GLogLevelFlags  log_level,
             const gchar    *message,
             gpointer        user_data)
{
    const gchar *log_level_string;
    const gchar *log_level_color;
    GDateTime   *date     = NULL;
    gchar       *time_fmt = NULL;

    /* Restrict log level */
    switch ((int) log_level)
    {
        case G_LOG_LEVEL_MESSAGE:
            if (log_verbosity < 1)
                return;
            break;

        case G_LOG_LEVEL_INFO:
            if (log_verbosity < 2)
                return;
            break;

        case G_LOG_LEVEL_DEBUG:
            if (log_verbosity < 3)
                return;
            break;

        default:
            break;
    }

    /* Get log level & color */
    if (isatty (1))
        log_level_string = log_level_to_string_with_color (log_level & G_LOG_LEVEL_MASK);
    else
        log_level_string = log_level_to_string (log_level & G_LOG_LEVEL_MASK);


    /* Time of log */
    date = g_date_time_new_now_local ();
    time_fmt = g_date_time_format (date, "%H:%M:%S");

    g_print ("[%s] [%s] %s: %s\n",
             time_fmt,
             log_level_string,
             log_domain,
             message);

    g_free (time_fmt);
    g_date_time_unref (date);
}

/**
 * boilerplate_log_init:
 *
 * Starts logging, sets the default log handler.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
void
boilerplate_log_init ()
{
    g_log_set_default_handler (log_handler, NULL);
}

/**
 * boilerplate_log_init:
 *
 * Increases the log verbose level by 1.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
void
boilerplate_log_increase_verbosity (void)
{
    log_verbosity++;
}

/**
 * boilerplate_log_init:
 *
 * Returns the current log level
 *
 * Returns: gint
 *
 * Since: 0.1.0
 */
gint
boilerplate_log_get_verbosity (void)
{
    return log_verbosity;
}

/**
 * boilerplate_log_init:
 * @level: The verbose level
 *
 * The verbose level that should be used
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
void
boilerplate_log_set_verbosity (gint level)
{
    log_verbosity = level;
}
