#define G_LOG_DOMAIN "boilerplate-window"

#include "config.h"

#include <gtk/gtk.h>

#include "boilerplate-application.h"
#include "boilerplate-window.h"

typedef struct {
    gint width;
    gint height;
    gint x;
    gint y;
    gboolean maximized;
    gboolean fullscreen;
} BoilerplateWindowState;

typedef struct
{
    GtkWidget *toolbar;
    GtkWidget *headerbar;
    GtkWidget *content;
} BoilerplateWindowPrivate;

struct _BoilerplateWindow {
    GtkApplicationWindow parent_instance;

    BoilerplateWindowState state;
};

#define STATE_TIMEOUT 100 /* ms */

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplateWindow, boilerplate_window, GTK_TYPE_APPLICATION_WINDOW)

/**
 * boilerplate_window_state_store:
 * @self: The window
 *
 * Saves the width, height, x, y, maximized and fullscreen values from the BoilerplateWindowState struct
 * into the gsettings: PACKAGE_ID ".window-state", eg. org.gnome.app.window-state
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_state_store (BoilerplateWindow *self)
{
    GSettings *settings = g_settings_new (PACKAGE_ID ".window-state");

    g_settings_set_int (settings, "width", self->state.width);
    g_settings_set_int (settings, "height", self->state.height);
    g_settings_set_int (settings, "x", self->state.x);
    g_settings_set_int (settings, "y", self->state.y);
    g_settings_set_boolean (settings, "maximized", self->state.maximized);
    g_settings_set_boolean (settings, "fullscreen", self->state.fullscreen);

    g_debug ("Saving WindowSate with: width->%i, height->%i, x->%i, y->%i",
             self->state.width,
             self->state.height,
             self->state.x,
             self->state.y);
}

/**
 * boilerplate_window_state_load:
 * @self: The window
 *
 * Loads the width, height, x, y, maximized and fullscreen values from gsettings
 * into the BoilerplateWindowState struct, located in the BoilerplateWindow struct
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_state_load (BoilerplateWindow *self)
{
    GSettings *settings = g_settings_new (PACKAGE_ID ".window-state");

    self->state.width = g_settings_get_int (settings, "width");
    self->state.height = g_settings_get_int (settings, "height");
    self->state.x = g_settings_get_int (settings, "x");
    self->state.y = g_settings_get_int (settings, "y");
    self->state.maximized = g_settings_get_boolean (settings, "maximized");
    self->state.fullscreen = g_settings_get_boolean (settings, "fullscreen");

    g_debug ("Loading WindowSate with: width->%i, height->%i, x->%i, y->%i",
             self->state.width,
             self->state.height,
             self->state.x,
             self->state.y);
}

/**
 * boilerplate_window_configure_event:
 * @widget: The widget (a.k.a the current window)
 * @GdkEventWindowState: The current state of the window
 *
 * Generats new information when a window size or position has changed.
 *
 * Returns: boolean, TRUE = event happened / FALSE = event did not happen
 *
 * Since: 0.1.0
 */
static gboolean
boilerplate_window_configure_event (GtkWidget         *widget,
                                    GdkEventConfigure *event)
{
    //TODO: Does this function cause memory issues?
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);
    gboolean retval;

    if (!(self->state.maximized || self->state.fullscreen)) {
        gtk_window_get_size (GTK_WINDOW (widget),
                             &self->state.width,
                             &self->state.height);

        //TODO: Make saving postion x and y work in Wayland?
        gtk_window_get_position (GTK_WINDOW (widget),
                                 &self->state.x,
                                 &self->state.y);
    }

    retval = GTK_WIDGET_CLASS (boilerplate_window_parent_class)->configure_event (widget, event);

    return retval;
}

/**
 * boilerplate_window_destroy:
 * @widget: The widget (a.k.a the current window)
 * @GdkEventWindowState: The current state of the window
 *
 * Used to set the current window state, a.k.a if the window is maximized or fullscreen
 *
 * Returns: boolean, TRUE = event happened / FALSE = event did not happen
 *
 * Since: 0.1.0
 */
static gboolean
boilerplate_window_window_state_event (GtkWidget           *widget,
                                       GdkEventWindowState *event)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);
    gboolean retval;

    self->state.maximized  = (event->new_window_state & GDK_WINDOW_STATE_MAXIMIZED) ? 1 : 0;
    self->state.fullscreen = (event->new_window_state & GDK_WINDOW_STATE_FULLSCREEN) ? 1 : 0;

    retval = GTK_WIDGET_CLASS (boilerplate_window_parent_class)->window_state_event (widget, event);

    return retval;
}

/**
 * boilerplate_window_constructed:
 * @object: The object (a.k.a the widget_class used / this GtkApplicationWindow)
 *
 * Sets initional window settings
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_constructed (GObject *object)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (object);

    self->state.width      = -1;
    self->state.height     = -1;
    self->state.x          =  0;
    self->state.y          =  0;
    self->state.maximized  =  FALSE;
    self->state.fullscreen =  FALSE;

    boilerplate_window_state_load (self);

    gtk_window_set_default_size (GTK_WINDOW (self),
                                 self->state.width,
                                 self->state.height);

    gtk_window_set_position (GTK_WINDOW (self), GTK_WIN_POS_MOUSE);

    if (self->state.x && self->state.y)
        gtk_window_move (GTK_WINDOW (self),
                         self->state.x,
                         self->state.y);

    if (self->state.maximized)
        gtk_window_maximize (GTK_WINDOW (self));

    if (self->state.fullscreen)
        gtk_window_fullscreen (GTK_WINDOW (self));

    G_OBJECT_CLASS (boilerplate_window_parent_class)->constructed (object);
}

/**
 * boilerplate_window_destroy:
 * @widget: The widget (a.k.a the current window)
 *
 * Cleans-up anything todo with the window/widget then frees the widget.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_destroy (GtkWidget *widget)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);

    boilerplate_window_state_store (self);

    GTK_WIDGET_CLASS (boilerplate_window_parent_class)->destroy (widget);
}

/**
 * boilerplate_window_finalize:
 * @object: The window
 *
 * Does any last minute clean-ups when the class ends
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_finalize (GObject *object)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (object);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    g_object_unref (priv->toolbar);

    G_OBJECT_CLASS (boilerplate_window_parent_class)->finalize(object);
}

/**
 * boilerplate_window_init:
 * @self: The window
 *
 * Initializes the application window.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_init (BoilerplateWindow *self)
{
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);
    GtkBuilder *builder;

    /* Init window.ui template */
    gtk_widget_init_template (GTK_WIDGET (self));

    /* Set window title */
    gtk_window_set_title (GTK_WINDOW (self), PACKAGE_NAME);

    /* Adding headerbar/titlebar/toolbar to private struct */
    if (priv->toolbar == NULL) {
        builder = gtk_builder_new_from_resource (PACKAGE_RESOURCES "/ui/toolbar.ui");
        priv->toolbar = GTK_WIDGET (gtk_builder_get_object (builder, "toolbar"));

        /* Ref toolbar so we can move it between the headerbar and toolbar */
        g_object_ref (priv->toolbar);

        g_object_unref (builder);
    }

    /* Set default window decorations (a.k.a GtkHeaderBar) */
    gtk_window_set_titlebar (GTK_WINDOW (self), priv->headerbar);
    gtk_container_add (GTK_CONTAINER (priv->headerbar), priv->toolbar);
}

/**
 * boilerplate_window_class_init:
 * @class: The window class
 *
 * Initializes the various functions in the window and links the gtk_widget
 * template loaded from gresource.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_class_init (BoilerplateWindowClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

    object_class->finalize = boilerplate_window_finalize;
    object_class->constructed = boilerplate_window_constructed;

    widget_class->destroy = boilerplate_window_destroy;
    widget_class->configure_event = boilerplate_window_configure_event;
    widget_class->window_state_event = boilerplate_window_window_state_event;

    gtk_widget_class_set_template_from_resource (widget_class,
                                                 PACKAGE_RESOURCES "/ui/window.ui");

    gtk_widget_class_bind_template_child_private (widget_class, BoilerplateWindow, headerbar);
    gtk_widget_class_bind_template_child_private (widget_class, BoilerplateWindow, content);
}

/**
 * boilerplate_window_new:
 * @application: The application
 *
 * Creates a new GtkApplicationWindow.
 *
 * Returns: BoilerplateWindow -> GtkApplicationWindow
 *
 * Since: 0.1.0
 */
BoilerplateWindow *
boilerplate_window_new (BoilerplateApplication *application)
{
    return g_object_new (BOILERPLATE_TYPE_WINDOW,
                         "application", application,
                         "title", PACKAGE_NAME,
                         NULL);
}

